      subroutine stadj(uadj, vadj, phiadj, u, v, phi, uobs, vobs, &
                       phiobs, UObsStddev, VObsStddev, PhiObsStddev, n, m, nt, k)
!***********************************************************
!  in this subroutine the adjoint variables have the same
!  name  as the direct variables in the direct system
!  the adjoint code is derived by writing backward
!  the linear tangent model
!  the adjoint of each instruction is written step by step
!********************************************************
      implicit none

! size parameters
      integer, intent(in) :: n, m, nt, k

! state variables
      double precision, intent(in), dimension(n, m, nt) :: u, v, phi

! observation variables
      double precision, intent(in), dimension(n, m, nt) :: uobs, vobs, &
                                                             phiobs

! weighting variables
      double precision, intent(in) :: UObsStddev, VObsStddev, PhiObsStddev

! adjoint variables
      double precision, intent(out), dimension(n, m, nt) :: uadj, &
                                                              vadj, &
                                                              phiadj

! local variables
      integer :: i, j
      double precision :: dphival, dvval, duval, uval, vval, phival
      double precision :: costu0, costv0, costp0

      costu0 = 1.0
      costv0 = 1.0
      costp0 = 1.0

      do j=2,m-1
          do i=2,n-1
              call obs_operator(u(i,j,k),v(i,j,k),phi(i,j,k),uval,vval,phival)

              call obs_operator_adj(u(i,j,k),v(i,j,k),phi(i,j,k),duval,dvval,dphival)

              uadj(i,j,k)   = 2.*(uval-uobs(i,j,k))*duval/(UObsStddev**2)/costu0
              vadj(i,j,k)   = 2.*(vval-vobs(i,j,k))*dvval/(VObsStddev**2)/costv0
              phiadj(i,j,k) = 2.*(phival-phiobs(i,j,k))*dphival/(PhiObsStddev**2)/costp0
          enddo
      enddo

      return
      end
