program swe_enkf
    use shared
    implicit none

    double precision,DIMENSION(:,:,:,:),allocatable :: EnsembleU
    double precision,DIMENSION(:,:,:,:),allocatable :: EnsembleV
    double precision,DIMENSION(:,:,:,:),allocatable :: EnsemblePhi

    double precision,DIMENSION(:,:),allocatable :: data2DPhi
    double precision,DIMENSION(:,:),allocatable :: data2DU
    double precision,DIMENSION(:,:),allocatable :: data2DV

    ! A: N_{state} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: perturbationU
    double precision,DIMENSION(:,:),allocatable :: perturbationV
    double precision,DIMENSION(:,:),allocatable :: perturbationPhi

    ! D, Y: N_{obs} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: ensObsIncrementU
    double precision,DIMENSION(:,:),allocatable :: ensObsIncrementV
    double precision,DIMENSION(:,:),allocatable :: ensObsIncrementPhi

    ! HX, M: N_{obs} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: ensObsU
    double precision,DIMENSION(:,:),allocatable :: ensObsV
    double precision,DIMENSION(:,:),allocatable :: ensObsPhi

    ! \bar{X}: N_{state}
    double precision,DIMENSION(:),allocatable :: meanU
    double precision,DIMENSION(:),allocatable :: meanV
    double precision,DIMENSION(:),allocatable :: meanPhi

    ! \bar{HX}: N_{obs}
    double precision,DIMENSION(:),allocatable :: meanObsU
    double precision,DIMENSION(:),allocatable :: meanObsV
    double precision,DIMENSION(:),allocatable :: meanObsPhi

    ! HA: N_{obs} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: ensObsPerturbationU
    double precision,DIMENSION(:,:),allocatable :: ensObsPerturbationV
    double precision,DIMENSION(:,:),allocatable :: ensObsPerturbationPhi

    ! P: N_{ens} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: choleskyU
    double precision,DIMENSION(:,:),allocatable :: choleskyV
    double precision,DIMENSION(:,:),allocatable :: choleskyPhi

    ! Z: N_{ens} \times N_{ens}
    double precision,DIMENSION(:,:),allocatable :: UEnsEnsTmp
    double precision,DIMENSION(:,:),allocatable :: VEnsEnsTmp
    double precision,DIMENSION(:,:),allocatable :: PhiEnsEnsTmp

! local variables
    integer :: i, j, k, kpass, ierr, member, stateindex
    double precision :: uval, vval, phival

    character(len=64) :: rmseFileName, file_num

    call initialize()

    ! X: N_{ens} \times n \times m \times nt
    call alloc4(EnsembleU,NoMembers,n,m,nt,'EnsembleU')
    call alloc4(EnsembleV,NoMembers,n,m,nt,'EnsembleV')
    call alloc4(EnsemblePhi,NoMembers,n,m,nt,'EnsemblePhi')

    ! helper array: n-2 \times m-2
    call alloc2(data2DU,n-2,m-2,'data2DU')
    call alloc2(data2DV,n-2,m-2,'data2DV')
    call alloc2(data2DPhi,n-2,m-2,'data2DPhi')

    ! A: N_{state} \times N_{ens}
    call alloc2(perturbationU,ns,NoMembers,'perturbationU')
    call alloc2(perturbationV,ns,NoMembers,'perturbationV')
    call alloc2(perturbationPhi,ns,NoMembers,'perturbationPhi')

    ! D, Y: N_{obs} \times N_{ens}
    call alloc2(ensObsIncrementU,ns,NoMembers,'ensObsIncrementU')
    call alloc2(ensObsIncrementV,ns,NoMembers,'ensObsIncrementV')
    call alloc2(ensObsIncrementPhi,ns,NoMembers,'ensObsIncrementPhi')

    ! HX, M: N_{obs} \times N_{ens}
    call alloc2(ensObsU,nobs,NoMembers,'ensObsU')
    call alloc2(ensObsV,nobs,NoMembers,'ensObsV')
    call alloc2(ensObsPhi,nobs,NoMembers,'ensObsPhi')

    ! \bar{X}: N_{state}
    call alloc1(meanU,ns,'meanU')
    call alloc1(meanV,ns,'meanV')
    call alloc1(meanPhi,ns,'meanPhi')

    ! \bar{HX}: N_{obs}
    call alloc1(meanObsU,nobs,'meanObsU')
    call alloc1(meanObsV,nobs,'meanObsV')
    call alloc1(meanObsPhi,nobs,'meanObsPhi')

    ! HA: N_{obs} \times N_{ens}
    call alloc2(ensObsPerturbationU,nobs,NoMembers,'ensObsPerturbationU')
    call alloc2(ensObsPerturbationV,nobs,NoMembers,'ensObsPerturbationV')
    call alloc2(ensObsPerturbationPhi,nobs,NoMembers,'ensObsPerturbationPhi')

    ! P: N_{ens} \times N_{ens}
    call alloc2(choleskyU,NoMembers,NoMembers,'choleskyU')
    call alloc2(choleskyV,NoMembers,NoMembers,'choleskyV')
    call alloc2(choleskyPhi,NoMembers,NoMembers,'choleskyPhi')

    ! Z: N_{ens} \times N_{ens}
    call alloc2(UEnsEnsTmp,NoMembers,NoMembers,'UEnsEnsTmp')
    call alloc2(VEnsEnsTmp,NoMembers,NoMembers,'VEnsEnsTmp')
    call alloc2(PhiEnsEnsTmp,NoMembers,NoMembers,'PhiEnsEnsTmp')

    call readFiles()

    ! read in the ensemble from the SWM_out_XXXX files
    do member=1,NoMembers

        print *, 'Reading the perturbations for member', member

        write(file_num,'(I0.4)') member

        open(unit=12,file='SWM_out_' // file_num,status='unknown', &
             form='formatted')
        read(12,*) ((data2DU(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DV(i,j),i=1,n-2),j=1,m-2), &
                   ((data2DPhi(i,j),i=1,n-2),j=1,m-2)
        close(12)

        EnsembleU(member, 2:n-1, 2:m-1, 1) = data2DU(1:n-2, 1:m-2)
        EnsembleV(member, 2:n-1, 2:m-1, 1) = data2DV(1:n-2, 1:m-2)
        EnsemblePhi(member, 2:n-1, 2:m-1, 1) = data2DPhi(1:n-2, 1:m-2)
    end do

    ! calculate the mean of the ensemble at the initial time
    meanU = 0.
    meanV = 0.
    meanPhi = 0.

    ! collect values for the mean
    do member=1,NoMembers
	    do i=2,n-1
	        do j=2,m-1
	            stateindex = map_2to1(i-1,j-1,n-2)

	            meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, 1)
	            meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, 1)
	            meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, 1)
	        enddo
	    enddo
	enddo

    ! and normalized
	meanU = meanU / NoMembers
	meanV = meanV / NoMembers
	meanPhi = meanPhi / NoMembers

    ! set the initial estimate to the mean of the ensemble
    do i=2,n-1
        do j=2,m-1
            stateindex = map_2to1(i-1,j-1,n-2)

            u(i,j,1) = meanU(stateindex)
            v(i,j,1) = meanV(stateindex)
            phi(i,j,1) = meanPhi(stateindex)
        enddo
    enddo

    ! run the estimate forward in time to get the initial RMSE
    do k=1,nt
        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        call swe_model(u(1:n, 1:m, kpass:k), &
                       v(1:n, 1:m, kpass:k), &
                       phi(1:n, 1:m, kpass:k), &
                       n, m, nt, kpass, k, k)

        call calculateRmse(k)
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

    ! write the RMSE to file
81  format('RMSE-EnKF-Obsop',i1,'.txt')

    write(rmseFileName,81) ObsOpNum

    open(unit=84, file=rmseFileName, status='unknown', form='formatted')

    ! perform the data assimilation
    do k=1,nt

        if (k .eq. 1) then
            kpass = k
        else if (k .eq. 2) then
            kpass = k - 1
        else
            kpass = k - 2
        endif

        do member=1,NoMembers

            ! advance the model forward in time
	        call swe_model(EnsembleU(member, 1:n, 1:m, kpass:k), &
	                       EnsembleV(member, 1:n, 1:m, kpass:k), &
	                       EnsemblePhi(member, 1:n, 1:m, kpass:k), &
	                       n, m, nt, kpass, k, k)

            ! get the mean ensemble
	        meanU = 0.
            meanV = 0.
            meanPhi = 0.

            ! and the mean ensemble observation
	        meanObsU = 0.
            meanObsV = 0.
            meanObsPhi = 0.

            ! collect values for the means
            do i=2,n-1
                do j=2,m-1

		            stateindex = map_2to1(i-1,j-1,n-2)

		            meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)

                    call obs_operator(EnsembleU(member, i, j, k), &
                                      EnsembleV(member, i, j, k), &
                                      EnsemblePhi(member, i, j, k), &
                                      uval,vval,phival)

                    ensObsU(stateindex,member) = uval
                    ensObsV(stateindex,member) = vval
                    ensObsPhi(stateindex,member) = phival

                    ensObsIncrementU(stateindex,member) = Uobs(i, j, k) + &
                        UObsStddev*random_normal() - uval

                    ensObsIncrementV(stateindex,member) = Vobs(i, j, k) + &
                        VObsStddev*random_normal() - vval

                    ensObsIncrementPhi(stateindex,member) = Phiobs(i, j, k) + &
                        PhiObsStddev*random_normal() - phival

                    meanObsU(stateindex) = meanObsU(stateindex) + uval
                    meanObsV(stateindex) = meanObsV(stateindex) + vval
                    meanObsPhi(stateindex) = meanObsPhi(stateindex) + phival
                enddo
            enddo
	    enddo

        ! normalize the means
	    meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        meanObsU = meanObsU / NoMembers
        meanObsV = meanObsV / NoMembers
        meanObsPhi = meanObsPhi / NoMembers

        ! get the perturbation matrices for the state and observation
	    do member=1,NoMembers
            do i=2,n-1
	            do j=2,m-1
		            stateindex = map_2to1(i-1,j-1,n-2)

	                perturbationU(stateindex,member) = EnsembleU(member, i, j, k) - &
	                    meanU(stateindex)

		            perturbationV(stateindex,member) = EnsembleV(member, i, j, k) - &
                        meanV(stateindex)

		            perturbationPhi(stateindex,member) = EnsemblePhi(member, i, j, k)- &
                        meanPhi(stateindex)

                    ensObsPerturbationU(stateindex,member) = ensObsU(stateindex,member) - &
                        meanObsU(stateindex)

                    ensObsPerturbationV(stateindex,member) = ensObsV(stateindex,member) - &
                        meanObsV(stateindex)

                    ensObsPerturbationPhi(stateindex,member) = ensObsPhi(stateindex,member) - &
                        meanObsPhi(stateindex)
	            enddo
            enddo
	    enddo

        ! create the matrix to be decomposed I + HA^\trans R^{-1} \frac{1}{N_{ens} - 1} HA
	    UEnsEnsTmp = 0.
        VEnsEnsTmp = 0.
        PhiEnsEnsTmp = 0.

        ! create the I
	    do i=1,NoMembers
	       UEnsEnsTmp(i,i) = 1.
           VEnsEnsTmp(i,i) = 1.
           PhiEnsEnsTmp(i,i) = 1.
	    enddo

        ! and the rest of the matrix
	    choleskyU = UEnsEnsTmp + matmul(transpose(ensObsPerturbationU),ensObsPerturbationU) / &
	       ((NoMembers-1)*(UObsStddev**2))

        choleskyV = VEnsEnsTmp + matmul(transpose(ensObsPerturbationV),ensObsPerturbationV) / &
           ((NoMembers-1)*(VObsStddev**2))

        choleskyPhi = PhiEnsEnsTmp + matmul(transpose(ensObsPerturbationPhi),ensObsPerturbationPhi) / &
           ((NoMembers-1)*(PhiObsStddev**2))

        ! Cholesky decomposition
        call dpotrf('L', NoMembers, choleskyU, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky decomposition for U had error',ierr
            stop
        endif

        call dpotrf('L', NoMembers, choleskyV, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky decomposition for V had error',ierr
            stop
        endif

        call dpotrf('L', NoMembers, choleskyPhi, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky decomposition for Phi had error',ierr
            stop
        endif

        ! Create the matrix to be solved
        UEnsEnsTmp = matmul(transpose(ensObsPerturbationU),ensObsIncrementU)/(UObsStddev**2)
        VEnsEnsTmp = matmul(transpose(ensObsPerturbationV),ensObsIncrementV)/(VObsStddev**2)
        PhiEnsEnsTmp = matmul(transpose(ensObsPerturbationPhi),ensObsIncrementPhi)/(PhiObsStddev**2)

        ! Call the Cholesky solve
        call dpotrs('L', NoMembers, NoMembers, choleskyU, NoMembers, UEnsEnsTmp, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky solve for U had error',ierr
            stop
        endif

        call dpotrs('L', NoMembers, NoMembers, choleskyV, NoMembers, VEnsEnsTmp, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky solve for V had error',ierr
            stop
        endif

        call dpotrs('L', NoMembers, NoMembers, choleskyPhi, NoMembers, PhiEnsEnsTmp, NoMembers, ierr)
        if (ierr .ne. 0) then
            print *,'Error: the Cholesky solve for Phi had error',ierr
            stop
        endif

        ! Create the right-hand matrix of the update: M = R^{-1} \left[Y - \frac{1}{N_{ens}-1}(HA)W\right]
        ensObsU = (ensObsIncrementU - matmul(ensObsPerturbationU, UEnsEnsTmp)/(NoMembers - 1))/(UObsStddev**2)
        ensObsV = (ensObsIncrementV - matmul(ensObsPerturbationV, VEnsEnsTmp)/(NoMembers - 1))/(VObsStddev**2)
        ensObsPhi = (ensObsIncrementPhi - matmul(ensObsPerturbationPhi, PhiEnsEnsTmp)/(NoMembers - 1))/(PhiObsStddev**2)

        ! And get the ensemble update by (HA)^\trans M
        perturbationU = matmul(perturbationU,UEnsEnsTmp)/(NoMembers - 1)
        perturbationV = matmul(perturbationV,VEnsEnsTmp)/(NoMembers - 1)
        perturbationPhi = matmul(perturbationPhi,PhiEnsEnsTmp)/(NoMembers - 1)

        meanU = 0.
        meanV = 0.
        meanPhi = 0.

        ! update the ensemble while collecting values for the mean
        do member=1,NoMembers
            do i=2,n-1
                do j=2,m-1
                    stateindex = map_2to1(i-1,j-1,n-2)

                    EnsembleU(member, i, j, k) = EnsembleU(member, i, j, k) + &
                        perturbationU(stateindex,member)

                    EnsembleV(member, i, j, k) = EnsembleV(member, i, j, k) + &
                        perturbationV(stateindex,member)

                    EnsemblePhi(member, i, j, k) = EnsemblePhi(member, i, j, k) + &
                        perturbationPhi(stateindex,member)

                    meanU(stateindex) = meanU(stateindex) + EnsembleU(member, i, j, k)
                    meanV(stateindex) = meanV(stateindex) + EnsembleV(member, i, j, k)
                    meanPhi(stateindex) = meanPhi(stateindex) + EnsemblePhi(member, i, j, k)
                enddo
            enddo

             call set_boundary_conditions(k, m, n, EnsembleU(  member, :, :, k), &
                                                   EnsembleV(  member, :, :, k), &
                                                   EnsemblePhi(member, :, :, k))
        enddo

        ! normalize once again
        meanU = meanU / NoMembers
        meanV = meanV / NoMembers
        meanPhi = meanPhi / NoMembers

        ! and set the best estimate to the ensemble mean
        do i=2,n-1
	        do j=2,m-1
                stateindex = map_2to1(i-1,j-1,n-2)

	            u(i,j,k) = meanU(stateindex)
                v(i,j,k) = meanV(stateindex)
                phi(i,j,k) = meanPhi(stateindex)
            enddo
        enddo

        ! calculate the RMSE
        call calculateRmse(k)

        write (84,*) k, sqrt(urmse)/rmseroot, sqrt(vrmse)/rmseroot, sqrt(phirmse)/rmseroot

        print *, 'Completed iteration', k
    enddo

    print *, '----------------------------------------------------------------------------------------------------------'

    print *, 'Final RMSE is: '


    do k=1,nt
        call calculateRmse(k)
    enddo

    close(84)

    ! SWE clean-up
    call finalize()

    ! EnKF clean-up
	deallocate(EnsembleU)
	deallocate(EnsembleV)
	deallocate(EnsemblePhi)

	deallocate(data2DU)
	deallocate(data2DV)
	deallocate(data2DPhi)

	deallocate(perturbationU)
	deallocate(perturbationV)
	deallocate(perturbationPhi)

	deallocate(ensObsIncrementU)
	deallocate(ensObsIncrementV)
	deallocate(ensObsIncrementPhi)

	deallocate(ensObsU)
	deallocate(ensObsV)
	deallocate(ensObsPhi)

	deallocate(meanU)
	deallocate(meanV)
	deallocate(meanPhi)

	deallocate(meanObsU)
	deallocate(meanObsV)
	deallocate(meanObsPhi)

	deallocate(ensObsPerturbationU)
	deallocate(ensObsPerturbationV)
	deallocate(ensObsPerturbationPhi)

	deallocate(choleskyU)
	deallocate(choleskyV)
	deallocate(choleskyPhi)

	deallocate(UEnsEnsTmp)
	deallocate(VEnsEnsTmp)
	deallocate(PhiEnsEnsTmp)
end program
