program swe_getEnsemble
  use shared, only : mean, std

  implicit none

  integer::NX,NY

  character (len = *), parameter :: FIELD_NAME_U = "U"
  character (len = *), parameter :: FIELD_NAME_V = "V"
  character (len = *), parameter :: FIELD_NAME_Phi = "Phi"


  double precision,DIMENSION(:,:),allocatable :: data2DPhi 
  double precision,DIMENSION(:,:),allocatable :: data2DU
  double precision,DIMENSION(:,:),allocatable :: data2DV
  double precision,DIMENSION(:,:),allocatable :: U
  double precision,DIMENSION(:,:),allocatable :: V
  double precision,DIMENSION(:,:),allocatable :: Phi

  double precision,DIMENSION(:,:),allocatable :: CorrOpH

  double precision,DIMENSION(:,:,:),allocatable::EnsembleU
  double precision,DIMENSION(:,:,:),allocatable::EnsembleV
  double precision,DIMENSION(:,:,:),allocatable::EnsemblePhi

  integer::i,k,j,ierr

  double precision :: mn, sd

  character*80 file_num

  integer,dimension(8) :: values

  integer :: n, m, k_start, k_end, k_max
  namelist /sizedata/ n, m, k_start, k_end, k_max

  integer :: NoMembers
  double precision :: DX, DY, CorrH, CovInflation, UBackgroundStddev, VBackgroundStddev, PhiBackgroundStddev
  namelist /perturbation/ NoMembers, DX, DY, CorrH, CovInflation, UBackgroundStddev, VBackgroundStddev, PhiBackgroundStddev

  open(unit=42,file="namelist.swm",status='old')
  read(42,perturbation)
  close(42)

  if(NoMembers.le.1 .or. NoMembers.gt.400) then
     print*,'Improper ensemble size [2,400): ',NoMembers
     STOP
  end if

  if(DX.lt.0.0.or.DX.gt.1.0e+6) then
     print*,'Improper DX:',DX
     STOP
  end if

  if(DY.lt.0.0.or.DY.gt.1.0e+6) then
     print*,'Improper DY:',DY
     STOP
  end if

  if(CorrH.lt.0.0.or.CorrH.gt.1.0e+8) then
     print*,'Improper Correlation Length:',CorrH
     STOP
  end if

  if(UBackgroundStddev.lt.0.0.or.UBackgroundStddev.gt.1.0e+6) then
     print*,'Improper U stddev:',UBackgroundStddev
     STOP
  end if

  if(VBackgroundStddev.lt.0.0.or.VBackgroundStddev.gt.1.0e+6) then
     print*,'Improper V stddev:',VBackgroundStddev
     STOP
  end if

  if(PhiBackgroundStddev.lt.0.0.or.PhiBackgroundStddev.gt.1.0e+6) then
     print*,'Improper Phi variance:',PhiBackgroundStddev
     STOP
  end if

  print*,'Using the following options'
  print '(a,i3,a)',' - Ensemble size: ',NoMembers,' members'
  print '(a,f10.1,a,f10.1,a)',' - DX = ',DX,'km; DY = ',DY,'km'
  print '(a,f10.1,a)',' - Correlation distance H:',CorrH,'km'
  print '(a,f10.1)',' - U Stddev:',UBackgroundStddev
  print '(a,f10.1)',' - V Stddev:',VBackgroundStddev
  print '(a,f10.1)',' - Phi Stddev:',PhiBackgroundStddev
 
  print*,'Reading the domain dimensions from the namelist'
  open(unit=42,file="namelist.swm",status='old')
  read(42,sizedata)
  close(42)
  nx = n
  ny = m

  print '(a,i4)',' - NX:',NX
  print '(a,i4)',' - NY:',NY

  print*,'Allocating storage memory...'

  allocate(data2DPhi(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: data2DPhi"  
     STOP
  end if
  allocate(data2DU(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: data2DU"  
     STOP
  end if
  allocate(data2DV(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: data2DV" 
     STOP 
  end if
  allocate(U(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: U"  
     STOP
  end if
  allocate(V(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: V"  
     STOP
  end if
  allocate(Phi(NX,NY),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: Phi" 
     STOP 
  end if

  allocate(EnsembleU(NoMembers,NX-2,NY-2),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: EnsembleU" 
     STOP 
  end if
  allocate(EnsembleV(NoMembers,NX-2,NY-2),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: EnsembleV"  
     STOP
  end if
  allocate(EnsemblePhi(NoMembers,NX-2,NY-2),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: EnsemblePhi"  
     STOP
  end if

  ! call init_random_seed
  !call init_random_seed()

  ! read initial data
  open(unit=9,file='SWM_ini.txt',status='unknown', &
       form='formatted')
  read(9,*) ((U(i,j),i=1,nx),j=1,ny), &
            ((V(i,j),i=1,nx),j=1,ny), &
            ((Phi(i,j),i=1,nx),j=1,ny)
  close(9)

  print*,'Geopotential Phi:' 
  print*,'Allocate the horizontal operator'
  allocate(CorrOpH((NX-2)*(NY-2),(NX-2)*(NY-2)),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: CorrOpH"  
     STOP
  end if
 
  call ComputeCorrelations(NX,NY,NoMembers,CorrH,DX,DY,CorrOpH,'Phi',EnsemblePhi,PhiBackgroundStddev) 
  print*,'Deallocate the horizontal operator'
  deallocate(CorrOpH)
  
  print*,'Wind U-component:'
  print*,'Allocate the horizontal operator'
  allocate(CorrOpH((NX-2)*(NY-2),(NX-2)*(NY-2)),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: CorrOpH"  
     STOP
  end if
  call ComputeCorrelations(NX,NY,NoMembers,CorrH,DX,DY,CorrOpH,'U',EnsembleU,UBackgroundStddev) 
  print*,'Deallocate the horizontal operator'
  deallocate(CorrOpH)
  
  print*,'Wind V-component:'  
  print*,'Allocate the horizontal operator'
  allocate(CorrOpH((NX-2)*(NY-2),(NX-2)*(NY-2)),stat=ierr)
  if(ierr.ne.0) then
     write(*,*)"Allocation error: CorrOpH"  
     STOP
  end if
  call ComputeCorrelations(NX,NY,NoMembers,CorrH,DX,DY,CorrOpH,'V',EnsembleV,VBackgroundStddev)
  print*,'Deallocate the horizontal operator'
  deallocate(CorrOpH)
  
  print*,'Post processing statistics:'
  print*,'Phi ensemble: ',minval(EnsemblePhi(:,1:NX-2,1:NY-2)),maxval(EnsemblePhi(:,1:NX-2,1:NY-2))
  print*,'U-component ensemble: ',minval(EnsembleU(:,1:NX-2,1:NY-2)),maxval(EnsembleU(:,1:NX-2,1:NY-2))
  print*,'V-component ensemble: ',minval(EnsembleV(:,1:NX-2,1:NY-2)),maxval(EnsembleV(:,1:NX-2,1:NY-2))

  do k=1,NoMembers
     print*,'Writing the perturbations for member',k
     data2DU(2:NX-1,2:NY-1) = U(2:NX-1,2:NY-1) + reshape(EnsembleU(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))
     data2DV(2:NX-1,2:NY-1) = V(2:NX-1,2:NY-1) + reshape(EnsembleV(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))
     data2DPhi(2:NX-1,2:NY-1) = Phi(2:NX-1,2:NY-1) + reshape(EnsemblePhi(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))

     write(file_num,'(I0.4)') k
     open(unit=12,file='SWM_out_' // file_num,status='unknown', &
         form='formatted')
     write(12,*) ((data2DU(i,j),i=2,nx-1),j=2,ny-1), &
                 ((data2DV(i,j),i=2,nx-1),j=2,ny-1), &
                 ((data2DPhi(i,j),i=2,nx-1),j=2,ny-1)
     close(12)
  end do

  data2DU(:,:) = 0
  data2DV(:,:) = 0
  data2DPhi(:,:) = 0

  print *, 'Calculating the mean...'

  do k=1,NoMembers
     data2DU(2:NX-1,2:NY-1) = data2DU(2:NX-1,2:NY-1) + reshape(EnsembleU(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))
     data2DV(2:NX-1,2:NY-1) = data2DV(2:NX-1,2:NY-1) + reshape(EnsembleV(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))
     data2DPhi(2:NX-1,2:NY-1) = data2DPhi(2:NX-1,2:NY-1) + reshape(EnsemblePhi(k,1:NX-2,1:NY-2),(/NX-2,NY-2/))
  end do

  print*,'Post processing statistics 2:'

  data2DU = data2DU / NoMembers
  data2DV = data2DV / NoMembers
  data2DPhi = data2DPhi / NoMembers

    print*,'Phi ensemble: ',minval(data2DPhi(2:NX-1,2:NY-1)),maxval(data2DPhi(2:NX-1,2:NY-1))

    print*,'U-component ensemble: ',minval(data2DU(2:NX-1,2:NY-1)),maxval(data2DU(2:NX-1,2:NY-1))

    print*,'V-component ensemble: ',minval(data2DV(2:NX-1,2:NY-1)),maxval(data2DV(2:NX-1,2:NY-1))

  data2DU = U + data2DU
  data2DV = V + data2DV
  data2DPhi = Phi + data2DPhi

  print *, 'Writing the mean...'

  open(unit=12,file='SWM_out_mean',status='unknown',form='formatted')
  write(12,*) ((data2DU(i,j),i=2,NX-1),j=2,NY-1), &
              ((data2DV(i,j),i=2,NX-1),j=2,NY-1), &
              ((data2DPhi(i,j),i=2,NX-1),j=2,NY-1)
  close(12)

  print *, 'Program completed successfully.'
  
  deallocate(data2DPhi)
  deallocate(data2DU)
  deallocate(data2DV)
  deallocate(U)
  deallocate(V)
  deallocate(Phi)
  deallocate(EnsemblePhi)
  deallocate(EnsembleU)
  deallocate(EnsembleV)

end program swe_getEnsemble
!contains

subroutine ComputeCorrelations(NX,NY,NoMembers,CorrH,DX,DY,HorzCorr,FileSuffix,Ensemble,Stddev)
  use shared, only : map_1to2, random_normal, mean, std
  implicit none
  integer,intent(in)::NX,NY,NoMembers
  double precision,intent(in):: CorrH,DX,DY,Stddev
  double precision,dimension(NoMembers,NX-2,NY-2),intent(inout)::Ensemble
  double precision,DIMENSION((NX-2)*(NY-2),(NX-2)*(NY-2)) :: HorzCorr
  double precision,DIMENSION((NX-2)*(NY-2),(NX-2)*(NY-2)) :: HorzCorrSquared
  character(1),intent(in)  :: FileSuffix
  character (len = 256) :: fileName

  integer::NXY2

  double precision :: dist
  integer::ii,ij
  integer::i,k,j

  double precision::mn,sd
  double precision,dimension(NoMembers)::rand_vec

  integer::IterMembers,iNX,iNY

  integer::INFO

  double precision,DIMENSION((NX-2)*(NY-2)) :: OperandNXY

  integer,dimension(8) :: values
  double precision::CorrVStddev

  fileName  = 'HorzCorr' // FileSuffix // '.txt'

  NXY2=(NX-2)*(NY-2)
  HorzCorr=0.0

  ! Generate normally distributed noise centered on zero
  Ensemble=0.0 !zeros(NoMembers,NX,NY,NZ);
  do i=1,NX-2
     do j=1,NY-2
        do ii=1,NoMembers
           rand_vec(ii)=random_normal()
        end do

        Ensemble(1:NoMembers,i,j)=rand_vec
     end do
  end do

  ! correlate the noise, one dimension at a time
  call compute_horizontal_operator(NX,NY,DX,DY,CorrH,FileSuffix,HorzCorr)

  do IterMembers=1,NoMembers  
     operandNXY=reshape(Ensemble(IterMembers,1:NX-2,1:NY-2),(/NXY2/))
     operandNXY=MATMUL(HorzCorr,operandNXY)
     Ensemble(IterMembers,1:NX-2,1:NY-2)=Stddev*reshape(operandNXY,(/NX-2,NY-2/))
  end do

  close(12)
  open(unit=12,file=trim(fileName),status='unknown', &
      form='formatted')
  write(12,*) ((Stddev*HorzCorr(i,j),i=1,nxy2),j=1,nxy2)
  close(12)

  HorzCorrSquared = MATMUL(HorzCorr,HorzCorr)

  open(unit=12,file='BEC' // FileSuffix // '.txt',status='unknown', &
      form='formatted')
  write(12,*) ((HorzCorrSquared(i,j),i=1,nxy2),j=1,nxy2)
  close(12)

end subroutine ComputeCorrelations

subroutine compute_horizontal_operator(NX,NY,DX,DY,CorrH,FileSuffix,HorzCorr)
  use shared, only : map_1to2

  implicit none
  integer,intent(in)::NX,NY
  double precision,intent(in):: DX,DY
  double precision,intent(in):: CorrH
  double precision,DIMENSION((NX-2)*(NY-2),(NX-2)*(NY-2)),intent(out) :: HorzCorr
  character(len=4)::ch
  character(1),intent(in)  :: FileSuffix
  integer::NXY2

  double precision :: dist
  integer::i,j,ii,ij,ji,jj 

  integer,dimension(8) :: values

  NXY2=(NX-2)*(NY-2)

  HorzCorr=0.0

  do i=1,NXY2
     do j=1,NXY2
        call map_1to2(i,NX-2,ii,ij)
        call map_1to2(j,NX-2,ji,jj)
        dist=abs((DX*(ii-ji))**2+(DY*(ij-jj))**2);
        HorzCorr(i,j)=exp(-dist/CorrH**2)
        HorzCorr(j,i)=HorzCorr(i,j)
     end do
  end do

end subroutine compute_horizontal_operator
